#include "glassexe.h"

void printUsage(ostream& out) 
{
  out << "Usage:    glass <hfile> <mfile> [options]\n\n";
  out << "Options:\n\n";
  out << "\t-o <filename>\n";
  out << "\t\tPrint output to <filename> instead of STDOUT\n";
  out << "\t-c\n";
  out << "\t\tPrint Clustal W header in output\n";
  out << "\t-r\n";
  out << "\t\tOutput only aligning regions\n";
  out << "\t-p <filename>\n";
  out << "\t\tUse parameters specified in <filename>\n";
  out << endl;
}


void parseFromCmdLine(int argc, char* argv[], GlassExeRegistry& registry) 
{
  // Parse the optional arguments.
  int c = 0;
  while((c = getopt(argc, argv, "o:crp:")) != -1) {
    switch(c) {
    case 'o':
      registry.outputFileName = optarg;
      break;
    case 'c':
      registry.useClustalHeader = true;
      break;
    case 'r':
      registry.outputRegions = true;
      break;
    case 'p':
      registry.readParameters = true;
      registry.parameterFileName = optarg;
      break;
    case '?':
      cerr << "Unknown option -" << optopt << "." << endl;
      exit(EXIT_FAILURE);
    default:
      abort();
    }
  }
  // Parse the required arguments.
  if(optind + 2 == argc) {
    registry.humanFileName = string(argv[optind]);
    registry.mouseFileName = string(argv[optind+1]);
  }
  else {
    printUsage(cerr);
    exit(EXIT_FAILURE);
  }
}


void outputResult(ostream& out, vector<PairwiseAlignment>& img,
		  GlassExeRegistry& registry) {
  if(registry.useClustalHeader) {
      out << registry.clustalHeader;
  }
  else {
      out << registry.glassHeader;
  }
  if(!registry.outputRegions) {
    PairwiseAlignmentIO::writeClustal(out, img[0]);
  }
  else {
    PairwiseAlignmentIO::writeRegions(out, img);
  }
}


int main(int argc, char* argv[]) 
{
  GlassExeRegistry registry;
  parseFromCmdLine(argc, argv, registry);

  // Read the first sequence file. (The human/mouse names are arbitrary).
  ifstream hfile(registry.humanFileName.data());
  if(!hfile || hfile.bad()) {
    cerr << "File not found: " << registry.humanFileName << endl;
    exit(EXIT_FAILURE);
  }
  // Read the second sequence file.
  ifstream mfile(registry.mouseFileName.data());
  if(!mfile || mfile.bad()) {
    cerr << "File not found: " << registry.mouseFileName << endl;
    exit(EXIT_FAILURE);
  }
  
  Sequence hseq = SequenceIO::readFasta(hfile);
  Sequence mseq = SequenceIO::readFasta(mfile);  

  Glass glass;

  // Read parameters from file if -p swith is used.
  if(registry.readParameters) {
    ifstream pfile(registry.parameterFileName.data());
    if(!pfile || pfile.bad()) {
      cerr << "File not found: " << registry.parameterFileName << endl;
      exit(EXIT_FAILURE);
    }
    glass.setParamsFromFile(pfile);
  }

  // Compute the GLASS alignment.
  vector<PairwiseAlignment> img;
  if(registry.outputRegions) {
    img = glass.alignRegions(hseq, mseq);
  }
  else {
    img.push_back(glass.align(hseq, mseq));
  }

  // Write the output to file or stdout if not filename was provided.
  if(registry.outputFileName != "") {
    ofstream fout(registry.outputFileName.c_str());
    if(!fout || fout.bad()) {
      cerr << "Cannot create file: " << registry.outputFileName << endl;
      exit(EXIT_FAILURE);
    }
    outputResult(fout, img, registry);
  }
  else {
    outputResult(cout, img, registry);
  }

  exit(EXIT_SUCCESS);
}

